/* GtkNumberEntry - widget for gtk+
 * Copyright (C) 2001 Masami Ogoshi <ogochan@linux.or.jp>
 *
 * GtkNumberEntry widget by Masami Ogoshi
 * Based on GtkItemEntry widget
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
#define	DEBUG
#define	TRACE
*/
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <numeric.h>
#include <gdk/gdkkeysyms.h>
#include <gdk/gdki18n.h>
#include <gtk/gtkmain.h>
#include <gtk/gtkselection.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkentry.h>
#include "gtknumberentry.h"
#include "debug.h"

#define MIN_ENTRY_WIDTH  150
#define DRAW_TIMEOUT     20
#define INNER_BORDER     2

enum {
  ARG_0,
  ARG_FORMAT
};


static void gtk_number_entry_class_init   (GtkNumberEntryClass     *klass);
static void gtk_number_entry_init         (GtkNumberEntry          *entry);
static void gtk_number_entry_set_arg   	  (GtkObject         *object,
					   GtkArg            *arg,
					   guint              arg_id);
static void gtk_number_entry_get_arg	  (GtkObject         *object,
					   GtkArg            *arg,
					   guint              arg_id);
static void gtk_entry_set_position_from_editable 
                                          (GtkEditable *editable,
			                   gint      position);

static void gtk_entry_realize             (GtkWidget         *widget);
static void gtk_entry_size_allocate       (GtkWidget         *widget,
					   GtkAllocation     *allocation);
static void gtk_entry_make_backing_pixmap (GtkEntry *entry,
					   gint width, gint height);
static void gtk_entry_draw                (GtkWidget         *widget,
					   GdkRectangle      *area);
static gint gtk_entry_expose              (GtkWidget         *widget,
					   GdkEventExpose    *event);
static gint gtk_entry_button_press        (GtkWidget         *widget,
					   GdkEventButton    *event);
static gint gtk_entry_button_release      (GtkWidget         *widget,
					   GdkEventButton    *event);
static gint gtk_entry_motion_notify       (GtkWidget         *widget,
					   GdkEventMotion    *event);
static gint gtk_entry_key_press           (GtkWidget         *widget,
					   GdkEventKey       *event);
static void gtk_entry_draw_text           (GtkEntry          *entry);
static void gtk_entry_draw_cursor         (GtkEntry          *entry);
static void gtk_entry_draw_cursor_on_drawable
					  (GtkEntry          *entry,
					   GdkDrawable       *drawable);
static void gtk_entry_queue_draw          (GtkEntry          *entry);
static gint gtk_entry_timer               (gpointer           data);
static gint gtk_entry_position            (GtkEntry          *entry,
					   gint               x);
static void entry_adjust_scroll           (GtkEntry          *entry);
static void gtk_entry_grow_text           (GtkEntry          *entry);
static void gtk_entry_insert_text         (GtkEditable       *editable,
					   const gchar       *new_text,
					   gint               new_text_length,
					   gint              *position);
static void gtk_entry_delete_text         (GtkEditable       *editable,
					   gint               start_pos,
					   gint               end_pos);
/* To be removed */
static void gtk_move_forward_word         (GtkEntry          *entry);
static void gtk_move_backward_word        (GtkEntry          *entry);
static void gtk_select_word               (GtkEntry          *entry,
					   guint32            time);
static void gtk_select_line               (GtkEntry          *entry,
					   guint32            time);


static void gtk_entry_set_selection       (GtkEditable       *editable,
					   gint               start,
					   gint               end);

static gint gtk_entry_find_position       (GtkEntry          *entry, 
					   gint               position);
static void gtk_entry_set_position_from_editable (GtkEditable *editable,
						  gint         position);

static GtkWidgetClass *parent_class = NULL;
static GdkAtom ctext_atom = GDK_NONE;

static const GtkTextFunction control_keys[26] =
{
  NULL,												/* a */
  NULL,												/* b */
  (GtkTextFunction)gtk_editable_copy_clipboard,   /* c */
  NULL,												/* d */
  NULL,												/* e */
  NULL,												/* f */
  NULL,                                           /* g */
  NULL,												/* h */
  NULL,                                           /* i */
  NULL,                                           /* j */
  NULL,												/* k */
  NULL,                                           /* l */
  NULL,                                           /* m */
  NULL,                                           /* n */
  NULL,                                           /* o */
  NULL,                                           /* p */
  NULL,                                           /* q */
  NULL,                                           /* r */
  NULL,                                           /* s */
  NULL,                                           /* t */
  NULL,												/* u */
  (GtkTextFunction)gtk_editable_paste_clipboard,  /* v */
  NULL,												/* w */
  (GtkTextFunction)gtk_editable_cut_clipboard,    /* x */
  NULL,                                           /* y */
  NULL,                                           /* z */
};

static const GtkTextFunction alt_keys[26] =
{
  NULL,                                           /* a */
  (GtkTextFunction)gtk_move_backward_word,        /* b */
  NULL,                                           /* c */
  NULL,												/* d */
  NULL,                                           /* e */
  (GtkTextFunction)gtk_move_forward_word,         /* f */
  NULL,                                           /* g */
  NULL,                                           /* h */
  NULL,                                           /* i */
  NULL,                                           /* j */
  NULL,                                           /* k */
  NULL,                                           /* l */
  NULL,                                           /* m */
  NULL,                                           /* n */
  NULL,                                           /* o */
  NULL,                                           /* p */
  NULL,                                           /* q */
  NULL,                                           /* r */
  NULL,                                           /* s */
  NULL,                                           /* t */
  NULL,                                           /* u */
  NULL,                                           /* v */
  NULL,                                           /* w */
  NULL,                                           /* x */
  NULL,                                           /* y */
  NULL,                                           /* z */
};


GtkType
gtk_number_entry_get_type (void)
{
  static GtkType entry_type = 0;

  if (!entry_type)
    {
      static const GtkTypeInfo entry_info =
      {
	"GtkNumberEntry",
	sizeof (GtkNumberEntry),
	sizeof (GtkNumberEntryClass),
	(GtkClassInitFunc) gtk_number_entry_class_init,
	(GtkObjectInitFunc) gtk_number_entry_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL,
      };

      entry_type = gtk_type_unique (GTK_TYPE_ENTRY, &entry_info);
    }

  return entry_type;
}

static void
gtk_number_entry_class_init (GtkNumberEntryClass *klass)
{
	GtkObjectClass *object_class;
	GtkWidgetClass *widget_class;
	GtkEditableClass *editable_class;

dbgmsg(">gtk_number_entry_class_init");
	object_class = (GtkObjectClass*) klass;
	widget_class = (GtkWidgetClass*) klass;
	editable_class = (GtkEditableClass*) klass;
	parent_class = gtk_type_class (GTK_TYPE_ENTRY);

	gtk_object_add_arg_type ("GtkNumberEntry::format", GTK_TYPE_STRING, GTK_ARG_READWRITE, ARG_FORMAT);

	object_class->set_arg = gtk_number_entry_set_arg;
	object_class->get_arg = gtk_number_entry_get_arg;

	widget_class->realize = gtk_entry_realize;
	widget_class->size_allocate = gtk_entry_size_allocate;
	widget_class->draw = gtk_entry_draw;
	widget_class->expose_event = gtk_entry_expose;
	widget_class->button_press_event = gtk_entry_button_press;
	widget_class->button_release_event = gtk_entry_button_release;
	widget_class->motion_notify_event = gtk_entry_motion_notify;
	widget_class->key_press_event = gtk_entry_key_press;

	editable_class->insert_text = gtk_entry_insert_text;
	editable_class->delete_text = gtk_entry_delete_text;
	editable_class->changed = (void (*)(GtkEditable *)) entry_adjust_scroll;

	editable_class->move_cursor = NULL;
	editable_class->move_word = NULL;
	editable_class->move_to_column = NULL;

	editable_class->kill_char = NULL;
	editable_class->kill_word = NULL;
	editable_class->kill_line = NULL;

	editable_class->set_selection = gtk_entry_set_selection;
	editable_class->set_position = gtk_entry_set_position_from_editable;
dbgmsg("<gtk_number_entry_class_init");
}


static void
gtk_number_entry_set_arg (
	GtkObject	*object,
	GtkArg		*arg,
	guint		arg_id)
{
	GtkNumberEntry	*entry;
  
	entry = GTK_NUMBER_ENTRY (object);
  
	switch	(arg_id)	{
	  case	ARG_FORMAT:
		gtk_number_entry_set_format(entry, GTK_VALUE_STRING (*arg));
		break;
	  default:
		break;
    }
}

static void
gtk_number_entry_get_arg (
	GtkObject	*object,
	GtkArg		*arg,
	guint		arg_id)
{
	GtkNumberEntry	*entry;
  
	entry = GTK_NUMBER_ENTRY (object);
  
	switch (arg_id) {
	  case ARG_FORMAT:
		GTK_VALUE_STRING (*arg) = g_strdup (entry->format);
		break;
	  default:
		arg->type = GTK_TYPE_INVALID;
		break;
    }
}

static void
gtk_number_entry_init (GtkNumberEntry *number_entry)
{
	number_entry->format = NULL;
	number_entry->scale = 0;
	number_entry->value = NULL;
}

extern	GtkWidget*
gtk_number_entry_new ()
{
	GtkNumberEntry	*number_entry;

	number_entry = gtk_type_new (gtk_number_entry_get_type ());

	return	GTK_WIDGET(number_entry);
}

extern	GtkWidget*
gtk_number_entry_new_with_max_length (
	guint16 max)
{
	GtkNumberEntry *number_entry;

	number_entry = gtk_type_new (gtk_number_entry_get_type ());
	GTK_ENTRY (number_entry)->text_max_length = max;

	return GTK_WIDGET (number_entry);
}

extern	GtkWidget*
gtk_number_entry_new_with_format(
	char	*fmt)
{
	GtkNumberEntry *number_entry;

	number_entry = gtk_type_new (gtk_number_entry_get_type ());
	if		(  number_entry->format  !=  NULL  ) {
		g_free(number_entry->format);
	}
	number_entry->format = g_strdup(fmt);

	return GTK_WIDGET (number_entry);
}

static	void
gtk_number_entry_redraw (
	GtkNumberEntry	*number_entry)
{
	gint tmp_pos;

	GtkEditable *editable;
	GtkEntry *entry;
	char	buf[64];

dbgmsg(">gtk_number_entry_redraw");
	g_return_if_fail (number_entry != NULL);
	g_return_if_fail (GTK_IS_NUMBER_ENTRY (number_entry));

	editable = GTK_EDITABLE (number_entry);
	entry = GTK_ENTRY(number_entry);

	gtk_entry_delete_text (GTK_EDITABLE(entry), 0, entry->text_length);

	NumericFormat(buf,number_entry->format,number_entry->value);

	tmp_pos = 0;
	gtk_editable_insert_text (editable, buf, strlen (buf), &tmp_pos);
	editable->current_pos = tmp_pos;

	editable->selection_start_pos = 0;
	editable->selection_end_pos = 0;

	if (GTK_WIDGET_DRAWABLE (entry))
		gtk_entry_draw_text (entry);
dbgmsg("<gtk_number_entry_redraw");
}

extern	void
gtk_number_entry_set_value (
	GtkNumberEntry	*number_entry,
	Numeric			value)
{
	int		pr
	,		sc;

dbgmsg(">gtk_number_entry_set_value");
	g_return_if_fail (number_entry != NULL);
	g_return_if_fail (GTK_IS_NUMBER_ENTRY (number_entry));

	if		(  number_entry->value  !=  value  ) {
		NumericFormatToPrecision(number_entry->format,&pr,&sc);
		number_entry->value = NumericRescale(value,pr+1,sc);
		number_entry->scale = 0;
		number_entry->expo = 0;
	} else {
		dbgmsg("not set value");
	}
	gtk_number_entry_redraw(number_entry);
dbgmsg("<gtk_number_entry_set_value");
}

extern	Numeric
gtk_number_entry_get_value (
	GtkNumberEntry	*entry)
{
	int		pr
	,		sc;
	
dbgmsg(">gtk_number_entry_get_value");
	g_return_val_if_fail (entry != NULL, 0);
	g_return_val_if_fail (GTK_IS_NUMBER_ENTRY (entry), 0);

	NumericFormatToPrecision(entry->format,&pr,&sc);
	
dbgmsg("<gtk_number_entry_get_value");
	return	(NumericRescale(entry->value,pr,sc));
}

static void
gtk_entry_set_position_from_editable (GtkEditable *editable,
			              gint      position)
{
  GtkEntry *entry;

  g_return_if_fail (editable != NULL);
  g_return_if_fail (GTK_IS_EDITABLE (editable));

  entry = GTK_ENTRY(editable);

  if ((position == -1) || (position > entry->text_length))
    GTK_EDITABLE(entry)->current_pos = entry->text_length;
  else
    GTK_EDITABLE(entry)->current_pos = position;
  entry_adjust_scroll (entry);
}

static GdkWChar
gtk_entry_get_invisible_char (GtkEntry *entry)
{
  GdkWChar ch = 0;

  if (entry->use_wchar)
    gdk_mbstowcs (&ch, "*", 1);
  else
    ch = '*';

  return ch;
}

/* Recompute the x offsets of all characters in the buffer */
static void
gtk_entry_recompute_offsets (GtkEntry *entry)
{
  gint i;
  gint offset = 0;
  GtkEditable *editable = GTK_EDITABLE (entry);

  for (i=0; i<entry->text_length; i++)
    {
      GdkWChar ch;

      entry->char_offset[i] = offset;

      if (editable->visible)
	ch = entry->text[i];
      else
	ch = gtk_entry_get_invisible_char (entry);

      if (entry->use_wchar)
	offset += gdk_char_width_wc (GTK_WIDGET (entry)->style->font, ch);
      else
	offset += gdk_char_width (GTK_WIDGET (entry)->style->font, ch);
    }
  
  entry->char_offset[i] = offset;
}

static void
gtk_entry_realize (GtkWidget *widget)
{
  GtkEntry *entry;
  GtkEditable *editable;
  GtkRequisition requisition;
  GdkWindowAttr attributes;
  gint attributes_mask;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_ENTRY (widget));

  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
  entry = GTK_ENTRY (widget);
  editable = GTK_EDITABLE (widget);

  gtk_widget_get_child_requisition (widget, &requisition);
  
  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y + (widget->allocation.height -
					 requisition.height) / 2;
  attributes.width = widget->allocation.width;
  attributes.height = requisition.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);
  attributes.event_mask = gtk_widget_get_events (widget);
  attributes.event_mask |= (GDK_EXPOSURE_MASK |
			    GDK_BUTTON_PRESS_MASK |
			    GDK_BUTTON_RELEASE_MASK |
			    GDK_BUTTON1_MOTION_MASK |
			    GDK_BUTTON3_MOTION_MASK |
			    GDK_POINTER_MOTION_HINT_MASK |
			    GDK_ENTER_NOTIFY_MASK |
			    GDK_LEAVE_NOTIFY_MASK |
			    GDK_KEY_PRESS_MASK);
  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

  widget->window = gdk_window_new (gtk_widget_get_parent_window (widget), &attributes, attributes_mask);
  gdk_window_set_user_data (widget->window, entry);

  attributes.x = widget->style->klass->xthickness;
  attributes.y = widget->style->klass->ythickness;
  attributes.width = widget->allocation.width - attributes.x * 2;
  attributes.height = requisition.height - attributes.y * 2;
  attributes.cursor = entry->cursor = gdk_cursor_new (GDK_XTERM);
  attributes_mask |= GDK_WA_CURSOR;

  entry->text_area = gdk_window_new (widget->window, &attributes, attributes_mask);
  gdk_window_set_user_data (entry->text_area, entry);

  widget->style = gtk_style_attach (widget->style, widget->window);

  gdk_window_set_background (widget->window, &widget->style->base[GTK_WIDGET_STATE (widget)]);
  gdk_window_set_background (entry->text_area, &widget->style->base[GTK_WIDGET_STATE (widget)]);

  gdk_window_show (entry->text_area);

  if (editable->selection_start_pos != editable->selection_end_pos)
    gtk_editable_claim_selection (editable, TRUE, GDK_CURRENT_TIME);

  gtk_entry_recompute_offsets (entry);
}

static void
gtk_entry_size_allocate (
	GtkWidget		*widget,
	GtkAllocation	*allocation)
{
	GtkEntry *entry;
	GtkEditable *editable;

	g_return_if_fail (widget != NULL);
	g_return_if_fail (GTK_IS_ENTRY (widget));
	g_return_if_fail (allocation != NULL);

	widget->allocation = *allocation;
	entry = GTK_ENTRY (widget);
	editable = GTK_EDITABLE (widget);

	if (GTK_WIDGET_REALIZED (widget)) {
		/* We call gtk_widget_get_child_requisition, since we want (for
		 * backwards compatibility reasons) the realization here to
		 * be affected by the usize of the entry, if set
		 */
		GtkRequisition requisition;
		gtk_widget_get_child_requisition (widget, &requisition);
  
		gdk_window_move_resize (widget->window,
								allocation->x,
								allocation->y + (allocation->height - requisition.height) / 2,
								allocation->width, requisition.height);
		gdk_window_move_resize (entry->text_area,
								widget->style->klass->xthickness,
								widget->style->klass->ythickness,
								allocation->width - widget->style->klass->xthickness * 2,
								requisition.height - widget->style->klass->ythickness * 2);

		/* And make sure the cursor is on screen */
		entry_adjust_scroll (entry);
    }
}

static void
gtk_entry_draw (GtkWidget    *widget,
		GdkRectangle *area)
{
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_ENTRY (widget));
  g_return_if_fail (area != NULL);

  if (GTK_WIDGET_DRAWABLE (widget))
    {
      gtk_widget_draw_focus (widget);
      gtk_entry_draw_text (GTK_ENTRY (widget));
    }
}

static gint
gtk_entry_expose (GtkWidget      *widget,
		  GdkEventExpose *event)
{
	GtkEntry *entry;

	g_return_val_if_fail (widget != NULL, FALSE);
	g_return_val_if_fail (GTK_IS_ENTRY (widget), FALSE);
	g_return_val_if_fail (event != NULL, FALSE);

	entry = GTK_ENTRY (widget);

	if (widget->window == event->window)
		gtk_widget_draw_focus (widget);
	else if (entry->text_area == event->window)
		gtk_entry_draw_text (GTK_ENTRY (widget));

	return	FALSE;
}

static gint
gtk_entry_button_press (GtkWidget      *widget,
			GdkEventButton *event)
{
	GtkEntry *entry;
	GtkEditable *editable;
	GdkModifierType mods;
	gint tmp_pos;

	g_return_val_if_fail (widget != NULL, FALSE);
	g_return_val_if_fail (GTK_IS_ENTRY (widget), FALSE);
	g_return_val_if_fail (event != NULL, FALSE);

	if (ctext_atom == GDK_NONE)
		ctext_atom = gdk_atom_intern ("COMPOUND_TEXT", FALSE);

	entry = GTK_ENTRY (widget);
	editable = GTK_EDITABLE (widget);
  
	if (entry->button && (event->button != entry->button))
		return FALSE;

	gdk_window_get_pointer(widget->window, NULL, NULL, &mods);
	if (mods & GDK_BUTTON3_MASK)
		return FALSE;

	entry->button = event->button;
  
	if (!GTK_WIDGET_HAS_FOCUS (widget))
		gtk_widget_grab_focus (widget);

	if (event->button == 1)	{
		switch (event->type)	{
		  case GDK_BUTTON_PRESS:
			gtk_grab_add (widget);
			tmp_pos = gtk_entry_position (entry, event->x + entry->scroll_offset);
			/* Set it now, so we display things right. We'll unset it
			 * later if things don't work out */
			editable->has_selection = TRUE;
			gtk_entry_set_selection (editable, tmp_pos, tmp_pos);
			editable->current_pos = editable->selection_start_pos;
			break;
		  case GDK_2BUTTON_PRESS:
			gtk_select_word (entry, event->time);
			break;
		  case GDK_3BUTTON_PRESS:
			gtk_select_line (entry, event->time);
			break;
		  default:
			break;
		}
    } else
	if (event->type == GDK_BUTTON_PRESS)	{
		if ((event->button == 2) && editable->editable)	{
			if (editable->selection_start_pos == editable->selection_end_pos ||
				editable->has_selection)
				editable->current_pos = gtk_entry_position (entry, event->x + entry->scroll_offset);
			gtk_selection_convert (widget, GDK_SELECTION_PRIMARY,
								   ctext_atom, event->time);
		} else if (event->button == 4 || event->button == 5) {
			return TRUE;
		} else {
			gtk_grab_add (widget);

			tmp_pos = gtk_entry_position (entry, event->x + entry->scroll_offset);
			gtk_entry_set_selection (editable, tmp_pos, tmp_pos);
			editable->has_selection = FALSE;
			editable->current_pos = editable->selection_start_pos;

			if (gdk_selection_owner_get (GDK_SELECTION_PRIMARY) == widget->window)
				gtk_selection_owner_set (NULL, GDK_SELECTION_PRIMARY, event->time);
		}
    }

	return	FALSE;
}

static gint
gtk_entry_button_release (GtkWidget      *widget,
			  GdkEventButton *event)
{
	GtkEntry *entry;
	GtkEditable *editable;

	g_return_val_if_fail (widget != NULL, FALSE);
	g_return_val_if_fail (GTK_IS_ENTRY (widget), FALSE);
	g_return_val_if_fail (event != NULL, FALSE);

	entry = GTK_ENTRY (widget);
	editable = GTK_EDITABLE (widget);

	if (entry->button != event->button)
		return FALSE;

	entry->button = 0;
  
	if (event->button == 1)	{
		gtk_grab_remove (widget);

		editable->has_selection = FALSE;
		if (editable->selection_start_pos != editable->selection_end_pos)	{
			if (gtk_selection_owner_set (widget,
										 GDK_SELECTION_PRIMARY,
										 event->time))
				editable->has_selection = TRUE;
			else
				gtk_entry_queue_draw (entry);
		} else {
			if (gdk_selection_owner_get (GDK_SELECTION_PRIMARY) == widget->window)
				gtk_selection_owner_set (NULL, GDK_SELECTION_PRIMARY, event->time);
		}
    } else
	if (event->button == 3)	{
		gtk_grab_remove (widget);
	}

	return FALSE;
}

static gint
gtk_entry_motion_notify (GtkWidget      *widget,
			 GdkEventMotion *event)
{
	GtkEntry *entry;
	gint x;

	g_return_val_if_fail (widget != NULL, FALSE);
	g_return_val_if_fail (GTK_IS_ENTRY (widget), FALSE);
	g_return_val_if_fail (event != NULL, FALSE);

	entry = GTK_ENTRY (widget);

	if (entry->button == 0)
		return FALSE;

	x = event->x;
	if (event->is_hint || (entry->text_area != event->window))
		gdk_window_get_pointer (entry->text_area, &x, NULL, NULL);

	GTK_EDITABLE(entry)->selection_end_pos = gtk_entry_position (entry, x + entry->scroll_offset);
	GTK_EDITABLE(entry)->current_pos = GTK_EDITABLE(entry)->selection_end_pos;
	gtk_entry_queue_draw (entry);

	return FALSE;
}

static gint
gtk_entry_key_press (
	GtkWidget	*widget,
	GdkEventKey	*event)
{
	GtkNumberEntry *entry;
	GtkEditable *editable;

	gint return_val;
	gint key;
	guint initial_pos;
	gint extend_start;
	char	*p;
	int		i;
	int		minus;
	Numeric	value
	,		value2
	,		input
	,		small
	,		one;
	int		pr
	,		sc;

dbgmsg(">gtk_entry_key_press");
	g_return_val_if_fail (widget != NULL, FALSE);
	g_return_val_if_fail (GTK_IS_ENTRY (widget), FALSE);
	g_return_val_if_fail (event != NULL, FALSE);

	entry = GTK_NUMBER_ENTRY (widget);
	editable = GTK_EDITABLE (widget);
	return_val = FALSE;

	if(!GTK_WIDGET_HAS_FOCUS(widget)) return FALSE;

	if		(  editable->editable  ==  FALSE  )	{
		switch(event->keyval){
		  case GDK_Left:
		  case GDK_Right:
		  case GDK_Up:
		  case GDK_Down:
		  case GDK_Tab:
		  case GDK_Return:
			break;
		  default:
			return TRUE;     
		}
	}

	initial_pos = editable->current_pos;

	extend_start = FALSE;

	NumericFormatToPrecision(entry->format,&pr,&sc);
	switch(event->keyval)	{
	  case GDK_BackSpace:
	  case GDK_Clear:
	  case GDK_Home:
	  case GDK_End:
		return_val = TRUE;
		entry->scale = 0;
		entry->expo = 0;
		NumericFree(entry->value);
		entry->value = NumericInput("0",pr+1,sc);
		break;
	  case GDK_Insert:
		return_val = TRUE;
		if (event->state & GDK_SHIFT_MASK) {
			gtk_editable_paste_clipboard (editable);
		} else
		if (event->state & GDK_CONTROL_MASK)	{
			gtk_editable_copy_clipboard (editable);
		}
		break;
	  case GDK_Delete:
		return_val = TRUE;
		if (event->state & GDK_SHIFT_MASK)	{
			gtk_editable_cut_clipboard (editable);
		}
		break;
	  case GDK_Left:
	  case GDK_Right:
		return_val = TRUE;
		break;
	  case GDK_Shift_L:
	  case GDK_Shift_R:
	  case GDK_Tab:
	  case GDK_Escape:
	  case GDK_Up:
	  case GDK_Down:
		return (* GTK_WIDGET_CLASS (parent_class)->key_press_event) (widget, event);
	  case GDK_Return:
		return_val = TRUE;
		gtk_signal_emit_by_name (GTK_OBJECT (entry), "activate");
		break;
	  default:
		if	((event->keyval >= 0x20) && (event->keyval <= 0xFF))	{
			key = event->keyval;

			if (event->state & GDK_CONTROL_MASK)	{
				if ((key >= 'A') && (key <= 'Z'))
					key -= 'A' - 'a';

				if ((key >= 'a') && (key <= 'z') && control_keys[key - 'a'])	{
					(* control_keys[key - 'a']) (editable, event->time);
					return_val = TRUE;
				}
				break;
			} else
			if (event->state & GDK_MOD1_MASK)	{
				if ((key >= 'A') && (key <= 'Z'))
					key -= 'A' - 'a';

				if ((key >= 'a') && (key <= 'z') && alt_keys[key - 'a'])	{
					(* alt_keys[key - 'a']) (editable, event->time);
					return_val = TRUE;
				}
				break;
			}
		}
		if		(  event->length  >  0  ) {
			gtk_editable_delete_selection (editable);
			if		(  NumericNegative(entry->value)  ) {
				minus = -1;
				value = NumericUMinus(entry->value);
				NumericFree(entry->value);
				entry->value = value;
			} else {
				minus = 1;
			}
			for	( p = event->string , i = 0 ; i < event->length ; i ++, p ++ ) {
				if		(  *p  ==  '-'  ) {
					minus = ( minus < 0 ) ? 1 : -1;
				} else
				if		(  *p  ==  '.'  ) {
					if		(  strchr(entry->format,'.')  !=  NULL  ) {
						entry->scale = 1;
					}
				} else
				if		(	(  *p  >=  '0'  )
						&&	(  *p  <=  '9'  ) ) {
					input = NumericInputChar(*p,1,0);
					if		(  entry->scale  ==  0  ) {
						if		(  entry->expo  <  ( pr - sc )  )	{
							value = NumericShift(entry->value,1);
							NumericFree(entry->value);
							entry->value = NumericADD(value,input);
							NumericFree(value);
							entry->expo ++;
						}
					} else {
						if		(  entry->scale  <=  sc  ) {
							one = NumericInput("1",pr,sc);
							small = NumericShift(one,- entry->scale);
							entry->scale ++;
							NumericFree(one);
							value = NumericMUL(small,input);
							NumericFree(small);
							value2 = NumericADD(entry->value,value);
							NumericFree(entry->value);
							NumericFree(value);
							entry->value = value2;
						}
					}
					NumericFree(input);
				}
			}
			if		(  minus  <  0  ) {
				value = NumericUMinus(entry->value);
				NumericFree(entry->value);
				entry->value = value;
			}
			return_val = TRUE;
		}
		break;
    }

	/* since we emit signals from within the above code,
	 * the widget might already be destroyed or at least
	 * unrealized.
	 */
	if		(	(  GTK_WIDGET_REALIZED (editable)  )
			&&	(  return_val                      ) )	{
		editable->selection_start_pos = 0;
		editable->selection_end_pos = 0;

		gtk_editable_claim_selection (editable,
									  editable->selection_start_pos != editable->selection_end_pos,
									  event->time);

		gtk_number_entry_redraw(entry);
    }

dbgmsg("<gtk_entry_key_press");
	return return_val;
}


static void
gtk_entry_make_backing_pixmap (GtkEntry *entry, gint width, gint height)
{
  gint pixmap_width, pixmap_height;

  if (!entry->backing_pixmap)
    {
      /* allocate */
      entry->backing_pixmap = gdk_pixmap_new (entry->text_area,
					      width, height,
					      -1);
    }
  else
    {
      /* reallocate if sizes don't match */
      gdk_window_get_size (entry->backing_pixmap,
			   &pixmap_width, &pixmap_height);
      if ((pixmap_width != width) || (pixmap_height != height))
	{
	  gdk_pixmap_unref (entry->backing_pixmap);
	  entry->backing_pixmap = gdk_pixmap_new (entry->text_area,
						  width, height,
						  -1);
	}
    }
}

static	void
gtk_entry_draw_text(
	GtkEntry *entry)
{
	GtkWidget *widget;
	GtkEditable *editable;
	GtkStateType selected_state;
	gint start_pos;
	gint end_pos;
	gint start_xoffset;
	gint selection_start_pos;
	gint selection_end_pos;
	gint selection_start_xoffset;
	gint selection_end_xoffset;
	gint width, height;
	gint y;
	GdkDrawable *drawable;
	gint use_backing_pixmap;
	GdkWChar *stars;
	GdkWChar *toprint;

	g_return_if_fail (entry != NULL);
	g_return_if_fail (GTK_IS_ENTRY (entry));

	if (entry->timer)	{
		gtk_timeout_remove (entry->timer);
		entry->timer = 0;
    }

	entry_adjust_scroll (entry);

	if (GTK_WIDGET_DRAWABLE (entry))	{
		widget = GTK_WIDGET (entry);
		editable = GTK_EDITABLE (entry);

		if (!entry->text)	{	  
			gtk_paint_flat_box (widget->style, entry->text_area,
								GTK_WIDGET_STATE(widget), GTK_SHADOW_NONE,
								NULL, widget, "entry_bg", 
								0, 0, -1, -1);

			if (editable->editable)
				gtk_entry_draw_cursor (entry);
			return;
		}

		gdk_window_get_size (entry->text_area, &width, &height);

		/*
		  If the widget has focus, draw on a backing pixmap to avoid flickering
		  and copy it to the text_area.
		  Otherwise draw to text_area directly for better speed.
		  */
		use_backing_pixmap = GTK_WIDGET_HAS_FOCUS (widget) && (entry->text != NULL);
		if (use_backing_pixmap)	{
			gtk_entry_make_backing_pixmap (entry, width, height);
			drawable = entry->backing_pixmap;
		} else {
			drawable = entry->text_area;
		}
		gtk_paint_flat_box (widget->style, drawable, 
							GTK_WIDGET_STATE(widget), GTK_SHADOW_NONE,
							NULL, widget, "entry_bg", 
							0, 0, width, height);

		y = (height - (widget->style->font->ascent + widget->style->font->descent)) / 2;
		y += widget->style->font->ascent;

		start_pos = gtk_entry_find_position (entry, entry->scroll_offset);
		start_xoffset = entry->char_offset[start_pos] - entry->scroll_offset;

		end_pos = gtk_entry_find_position (entry, entry->scroll_offset + width);
		if (end_pos < entry->text_length)
			end_pos += 1;

		selected_state = GTK_STATE_SELECTED;
		if (!editable->has_selection)
			selected_state = GTK_STATE_ACTIVE;

		selection_start_pos = MIN (editable->selection_start_pos, editable->selection_end_pos);
		selection_end_pos = MAX (editable->selection_start_pos, editable->selection_end_pos);
      
		selection_start_pos = CLAMP (selection_start_pos, start_pos, end_pos);
		selection_end_pos = CLAMP (selection_end_pos, start_pos, end_pos);

		selection_start_xoffset = 
			entry->char_offset[selection_start_pos] - entry->scroll_offset;
		selection_end_xoffset = 
			entry->char_offset[selection_end_pos] -entry->scroll_offset;

		/* if editable->visible, print a bunch of stars.  If not, print the standard text. */
		if (editable->visible)	{
			toprint = entry->text + start_pos;
		} else {
			gint i;
	  
			stars = g_new (GdkWChar, end_pos - start_pos);
			for (i = 0; i < end_pos - start_pos; i++)
				stars[i] = '*';
			toprint = stars;
		}
      
		if (selection_start_pos > start_pos)
			gdk_draw_text_wc (drawable, widget->style->font,
							  widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
							  INNER_BORDER + start_xoffset, y,
							  toprint,
							  selection_start_pos - start_pos);
      
		if ((selection_end_pos >= start_pos) && 
			(selection_start_pos < end_pos) &&
			(selection_start_pos != selection_end_pos))	{
			gtk_paint_flat_box (widget->style, drawable, 
								selected_state, GTK_SHADOW_NONE,
								NULL, widget, "text",
								INNER_BORDER + selection_start_xoffset,
								INNER_BORDER,
								selection_end_xoffset - selection_start_xoffset,
								height - 2*INNER_BORDER);
			gdk_draw_text_wc (drawable, widget->style->font,
							  widget->style->fg_gc[selected_state],
							  INNER_BORDER + selection_start_xoffset, y,
							  toprint + selection_start_pos - start_pos,
							  selection_end_pos - selection_start_pos);
		}
       
		if (selection_end_pos < end_pos)
			gdk_draw_text_wc (drawable, widget->style->font,
							  widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
							  INNER_BORDER + selection_end_xoffset, y,
							  toprint + selection_end_pos - start_pos,
							  end_pos - selection_end_pos);
		/* free the space allocated for the stars if it's neccessary. */
		if (!editable->visible)
			g_free (toprint);

		if (editable->editable)
			gtk_entry_draw_cursor_on_drawable (entry, drawable);

		if (use_backing_pixmap)
			gdk_draw_pixmap(entry->text_area,
							widget->style->fg_gc[GTK_STATE_NORMAL],
							entry->backing_pixmap,
							0, 0, 0, 0, width, height);	  
    }
}

static void
gtk_entry_draw_cursor (GtkEntry *entry)
{
	g_return_if_fail (entry != NULL);
	g_return_if_fail (GTK_IS_ENTRY (entry));

	gtk_entry_draw_cursor_on_drawable (entry, entry->text_area);
}

static	void
gtk_entry_draw_cursor_on_drawable (
	GtkEntry	*entry,
	GdkDrawable	*drawable)
{
	GtkWidget *widget;
	GtkEditable *editable;
	gint xoffset;
	gint text_area_height;

	g_return_if_fail (entry != NULL);
	g_return_if_fail (GTK_IS_ENTRY (entry));

	if (GTK_WIDGET_DRAWABLE (entry))	{
		widget = GTK_WIDGET (entry);
		editable = GTK_EDITABLE (entry);

		xoffset = INNER_BORDER + entry->char_offset[editable->current_pos];
		xoffset -= entry->scroll_offset;

		gdk_window_get_size (entry->text_area, NULL, &text_area_height);

		if (GTK_WIDGET_HAS_FOCUS (widget) &&
			(editable->selection_start_pos == editable->selection_end_pos))		{
			gdk_draw_line (drawable, widget->style->fg_gc[GTK_STATE_NORMAL], 
						   xoffset, INNER_BORDER,
						   xoffset, text_area_height - INNER_BORDER);
		} else {
			gint yoffset = 
				(text_area_height - 
				 (widget->style->font->ascent + widget->style->font->descent)) / 2
				+ widget->style->font->ascent;

			gtk_paint_flat_box (widget->style, drawable,
								GTK_WIDGET_STATE(widget), GTK_SHADOW_NONE,
								NULL, widget, "entry_bg", 
								xoffset, INNER_BORDER, 
								1, text_area_height - INNER_BORDER);
	  
			/* Draw the character under the cursor again
			 */
			if ((editable->current_pos < entry->text_length) &&
				(editable->selection_start_pos == editable->selection_end_pos))	{
				gdk_draw_text_wc (drawable, widget->style->font,
								  widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
								  xoffset, yoffset, (entry->text + editable->current_pos) , 1);
			}
		}
    }
}

static void
gtk_entry_queue_draw (GtkEntry *entry)
{
  g_return_if_fail (entry != NULL);
  g_return_if_fail (GTK_IS_ENTRY (entry));
  g_return_if_fail (GTK_IS_NUMBER_ENTRY (entry));

  if (!entry->timer)
    entry->timer = gtk_timeout_add (DRAW_TIMEOUT, gtk_entry_timer, entry);
}

static gint
gtk_entry_timer (
	gpointer data)
{
	GtkEntry *entry;

	GDK_THREADS_ENTER ();

	entry = GTK_ENTRY (data);
	entry->timer = 0;
	gtk_entry_draw_text (entry);

	GDK_THREADS_LEAVE ();

	return FALSE;
}

static gint
gtk_entry_find_position (
	GtkEntry	*entry,
	gint		x)
{
	gint start = 0;
	gint end = entry->text_length;
	gint half;

	if (x <= 0)
		return 0;
	if (x >= entry->char_offset[end])
		return end;
  
  /* invariant - char_offset[start] <= x < char_offset[end] */

	while (start != end)	{
		half = (start+end)/2;
		if (half == start)
			return half;
		else if (entry->char_offset[half] <= x)
			start = half;
		else
			end = half;
    }

	return start;
}

static gint
gtk_entry_position (GtkEntry *entry,
		    gint      x)
{
  return gtk_entry_find_position(entry, x);
}


static void
entry_adjust_scroll (GtkEntry *entry)
{
	GtkNumberEntry *number_entry;
	gint	text_area_width
	,		text_area_height;
	gint	char_width
	,		text_width;

dbgmsg(">gtk_adjust_scroll");
	g_return_if_fail (entry != NULL);
	g_return_if_fail (GTK_IS_ENTRY (entry));
	g_return_if_fail (GTK_IS_NUMBER_ENTRY (entry));

	number_entry = GTK_NUMBER_ENTRY (entry);

	if	(!entry->text_area)	{
dbgmsg("<gtk_adjust_scroll(null)");
		return;
	}

	gdk_window_get_size (entry->text_area, &text_area_width, &text_area_height);
	text_area_width -= 2 * INNER_BORDER;

	char_width = gdk_char_width (GTK_WIDGET(number_entry)->style->font,(gchar)'X');

	entry->scroll_offset = 0 ;

    if		(  entry->text  !=  NULL  )	{
		text_width = gdk_text_width (GTK_WIDGET (number_entry)->style->font, gtk_entry_get_text(entry), strlen(gtk_entry_get_text(entry)));
 
		entry->scroll_offset=  -(text_area_width - text_width);
		if		(  entry->scroll_offset  >  0  ) {
			if		(number_entry->text_max_size != 0 &&
					 text_area_width + char_width <= number_entry->text_max_size)	{
				GTK_WIDGET(number_entry)->allocation.x = 
					GTK_WIDGET(number_entry)->allocation.x+
					GTK_WIDGET(number_entry)->allocation.width-
					(text_area_width+char_width);
				GTK_WIDGET(number_entry)->allocation.width = text_area_width + char_width;
				gtk_entry_size_allocate(GTK_WIDGET(number_entry),
										(GtkAllocation *)(&GTK_WIDGET(number_entry)->allocation));
				gtk_widget_queue_draw(GTK_WIDGET(number_entry));
			} else {
				entry->scroll_offset= -(text_area_width  
										- gdk_text_width (GTK_WIDGET (number_entry)->style->font, gtk_entry_get_text(entry), GTK_EDITABLE(entry)->current_pos));
				if(entry->scroll_offset < 0) entry->scroll_offset = 0;
			}
		}
    } else {
		entry->scroll_offset=0;
	}
dbgmsg("<gtk_adjust_scroll");
}

static void
gtk_entry_grow_text (GtkEntry *entry)
{
	gint previous_size;
	gint i;

	g_return_if_fail (entry != NULL);
	g_return_if_fail (GTK_IS_ENTRY (entry));
	g_return_if_fail (GTK_IS_NUMBER_ENTRY (entry));

	previous_size = entry->text_size;
	if (!entry->text_size)
		entry->text_size = 128;
	else
		entry->text_size *= 2;
	entry->text = g_realloc (entry->text, entry->text_size * sizeof(GdkWChar));
	entry->char_offset = g_realloc (entry->char_offset, 
									entry->text_size * sizeof(guint));

	if (entry->text_length == 0)	{	/* initial allocation */
		entry->char_offset[0] = 0;
    }

	for (i = previous_size; i < entry->text_size; i++)
		entry->text[i] = '\0';
}

static void
gtk_entry_insert_text (GtkEditable *editable,
		       const gchar *new_text,
		       gint         new_text_length,
		       gint        *position)
{
  GdkWChar *text;
  gint start_pos;
  gint end_pos;
  gint last_pos;
  gint max_length;
  gint i;

  guchar *new_text_nt;
  gint insertion_length;
  GdkWChar *insertion_text;
  
  GtkEntry *entry;
  GtkWidget *widget;
  
dbgmsg(">gtk_entry_insert_text");
  g_return_if_fail (editable != NULL);
  g_return_if_fail (GTK_IS_ENTRY (editable));

  entry = GTK_ENTRY (editable);
  widget = GTK_WIDGET (editable);

  if ((entry->text_length == 0) && (entry->use_wchar == FALSE))
    {
      if (!GTK_WIDGET_REALIZED (widget))
	gtk_widget_ensure_style (widget);
      if ((widget->style) && (widget->style->font->type == GDK_FONT_FONTSET))
	entry->use_wchar = TRUE;
    }

  if (new_text_length < 0)
    {
      new_text_nt = (gchar *)new_text;
      new_text_length = strlen (new_text);
      if (new_text_length <= 0) return;
    }
  else if (new_text_length == 0)
    {
      return;
    }
  else
    {
      /* make a null-terminated copy of new_text */
      new_text_nt = g_new (gchar, new_text_length + 1);
      memcpy (new_text_nt, new_text, new_text_length);
      new_text_nt[new_text_length] = 0;
    }
    
  /* The algorithms here will work as long as, the text size (a
   * multiple of 2), fits into a guint16 but we specify a shorter
   * maximum length so that if the user pastes a very long text, there
   * is not a long hang from the slow X_LOCALE functions.  */
 
  if (entry->text_max_length == 0)
    max_length = 2047;
  else
    max_length = MIN (2047, entry->text_max_length);

  /* Convert to wide characters */
  insertion_text = g_new (GdkWChar, new_text_length);
  if (entry->use_wchar)
    insertion_length = gdk_mbstowcs (insertion_text, new_text_nt,
				     new_text_length);
  else
    for (insertion_length=0; new_text_nt[insertion_length]; insertion_length++)
      insertion_text[insertion_length] = new_text_nt[insertion_length];
  if (new_text_nt != (guchar *)new_text)
    g_free (new_text_nt);

  /* Make sure we do not exceed the maximum size of the entry. */
  if (insertion_length + entry->text_length > max_length)
    insertion_length = max_length - entry->text_length;

  /* Don't insert anything, if there was nothing to insert. */
  if (insertion_length <= 0)
    {
      g_free(insertion_text);
      return;
    }

  /* Make sure we are inserting at integral character position */
  start_pos = *position;
  if (start_pos < 0)
    start_pos = 0;
  else if (start_pos > entry->text_length)
    start_pos = entry->text_length;

  end_pos = start_pos + insertion_length;
  last_pos = insertion_length + entry->text_length;

  if (editable->selection_start_pos >= *position)
    editable->selection_start_pos += insertion_length;
  if (editable->selection_end_pos >= *position)
    editable->selection_end_pos += insertion_length;

  while (last_pos >= entry->text_size)
    gtk_entry_grow_text (entry);

  text = entry->text;
  for (i = last_pos - 1; i >= end_pos; i--)
    text[i] = text[i- (end_pos - start_pos)];
  for (i = start_pos; i < end_pos; i++)
    text[i] = insertion_text[i - start_pos];
  g_free (insertion_text);

  /* Fix up the the character offsets */
  
  if (GTK_WIDGET_REALIZED (entry))
    {
      gint offset = 0;
      
      for (i = last_pos; i >= end_pos; i--)
	entry->char_offset[i] 
	  = entry->char_offset[i - insertion_length];
      
      for (i=start_pos; i<end_pos; i++)
	{
	  entry->char_offset[i] = entry->char_offset[start_pos] + offset;
	  if (entry->visible)
	    {
	      offset += gdk_char_width_wc (GTK_WIDGET (entry)->style->font,
					   entry->text[i]);
	    }
	  else
	    {
	      offset += gdk_char_width (GTK_WIDGET (entry)->style->font, '*');
	    }
	}
      for (i = end_pos; i <= last_pos; i++)
	entry->char_offset[i] += offset;
    }

  entry->text_length += insertion_length;
  *position = end_pos;

  entry->text_mb_dirty = 1;
  gtk_entry_queue_draw (entry);
dbgmsg("<gtk_entry_insert_text");
}

static void
gtk_entry_delete_text (GtkEditable *editable,
		       gint         start_pos,
		       gint         end_pos)
{
  GdkWChar *text;
  gint deletion_length;
  gint i;

  GtkEntry *entry;
  
dbgmsg(">gtk_entry_delete_text");
  g_return_if_fail (editable != NULL);
  g_return_if_fail (GTK_IS_ENTRY (editable));

  entry = GTK_ENTRY (editable);

  if (end_pos < 0)
    end_pos = entry->text_length;

  if (editable->selection_start_pos > start_pos)
    editable->selection_start_pos -= MIN(end_pos, editable->selection_start_pos) - start_pos;
  if (editable->selection_end_pos > start_pos)
    editable->selection_end_pos -= MIN(end_pos, editable->selection_end_pos) - start_pos;
  
  if ((start_pos < end_pos) &&
      (start_pos >= 0) &&
      (end_pos <= entry->text_length))
    {
      text = entry->text;
      deletion_length = end_pos - start_pos;

      /* Fix up the character offsets */
      if (GTK_WIDGET_REALIZED (entry))
	{
	  gint deletion_width = 
	    entry->char_offset[end_pos] - entry->char_offset[start_pos];

	  for (i = 0 ; i <= entry->text_length - end_pos; i++)
	    entry->char_offset[start_pos+i] = entry->char_offset[end_pos+i] - deletion_width;
	}

      for (i = end_pos; i < entry->text_length; i++)
        text[i - deletion_length] = text[i];

      for (i = entry->text_length - deletion_length; i < entry->text_length; i++)
        text[i] = '\0';

      entry->text_length -= deletion_length;
      editable->current_pos = start_pos;
    }

  entry->text_mb_dirty = 1;
  gtk_entry_queue_draw (entry);
dbgmsg("<gtk_entry_delete_text");
}

static void
gtk_move_forward_word (GtkEntry *entry)
{
  GtkEditable *editable;
  GdkWChar *text;
  gint i;

  editable = GTK_EDITABLE (entry);

  if (entry->text && (editable->current_pos < entry->text_length))
    {
      text = entry->text;
      i = editable->current_pos;
	  
      if ((entry->use_wchar) ? (!gdk_iswalnum (text[i])) : (!isalnum (text[i])))
	for (; i < entry->text_length; i++)
	  {
	    if ((entry->use_wchar) ? gdk_iswalnum (text[i]) : isalnum (text[i]))
	      break;
	  }

      for (; i < entry->text_length; i++)
	{
	  if ((entry->use_wchar) ? (!gdk_iswalnum (text[i])) : (!isalnum (text[i])))
	    break;
	}

      editable->current_pos = i;
    }
}

static void
gtk_move_backward_word (GtkEntry *entry)
{
  GtkEditable *editable;
  GdkWChar *text;
  gint i;

  editable = GTK_EDITABLE (entry);

  if (entry->text && editable->current_pos > 0)
    {
      text = entry->text;
      i = editable->current_pos - 1;
      if ((entry->use_wchar) ? (!gdk_iswalnum (text[i])) : (!isalnum (text[i])))
	for (; i >= 0; i--)
	  {
	    if ((entry->use_wchar) ? gdk_iswalnum (text[i]) : isalnum (text[i]))
	      break;
	  }
      for (; i >= 0; i--)
	{
	  if ((entry->use_wchar) ? (!gdk_iswalnum (text[i])) : (!isalnum (text[i])))
	    {
	      i++;
	      break;
	    }
	}
	  
      if (i < 0)
	i = 0;
	  
      editable->current_pos = i;
    }
}

static void
gtk_select_word (GtkEntry *entry,
		 guint32   time)
{
  GtkEditable *editable;
  gint start_pos;
  gint end_pos;

  editable = GTK_EDITABLE (entry);

  gtk_move_backward_word (entry);
  start_pos = editable->current_pos;

  gtk_move_forward_word (entry);
  end_pos = editable->current_pos;

  editable->has_selection = TRUE;
  gtk_entry_set_selection (editable, start_pos, end_pos);
  gtk_editable_claim_selection (editable, start_pos != end_pos, time);
}

static void
gtk_select_line (GtkEntry *entry,
		 guint32   time)
{
  GtkEditable *editable;

  editable = GTK_EDITABLE (entry);

  editable->has_selection = TRUE;
  gtk_entry_set_selection (editable, 0, entry->text_length);
  gtk_editable_claim_selection (editable, entry->text_length != 0, time);

  editable->current_pos = editable->selection_end_pos;
}

static void 
gtk_entry_set_selection (GtkEditable       *editable,
			 gint               start,
			 gint               end)
{
  g_return_if_fail (editable != NULL);
  g_return_if_fail (GTK_IS_ENTRY (editable));

  if (end < 0)
    end = GTK_ENTRY (editable)->text_length;
  
  editable->selection_start_pos = start;
  editable->selection_end_pos = end;

  gtk_entry_queue_draw (GTK_ENTRY (editable));
}

extern	void
gtk_number_entry_set_format(
	GtkNumberEntry	*number_entry,
	const gchar		*fmt)
{
	g_return_if_fail (number_entry != NULL);
	g_return_if_fail (GTK_IS_NUMBER_ENTRY (number_entry));

	if		(  number_entry->format  !=  NULL  ) {
		g_free(number_entry->format);
	}
	number_entry->format = g_strdup(fmt);
}
