/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball, Josh MacDonald
 * Copyright (C) 1997-1998 Jay Painter <jpaint@serv.net><jpaint@gimp.org>
 * Copyright (C) 2001 AXE, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Based on gtk/gtkclist.h
 */

/*
 * Modified by the GTK+ Team and others 1997-1999.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
 */

#ifndef __GTK_PANDA_CLIST_H__
#define __GTK_PANDA_CLIST_H__

#include <gdk/gdk.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkalignment.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtkhscrollbar.h>
#include <gtk/gtkvscrollbar.h>
#include <gtk/gtkenums.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/* clist flags */
enum {
  GTK_PANDA_CLIST_IN_DRAG             = 1 <<  0,
  GTK_PANDA_CLIST_ROW_HEIGHT_SET      = 1 <<  1,
  GTK_PANDA_CLIST_SHOW_TITLES         = 1 <<  2,
  GTK_PANDA_CLIST_CHILD_HAS_FOCUS     = 1 <<  3,
  GTK_PANDA_CLIST_ADD_MODE            = 1 <<  4,
  GTK_PANDA_CLIST_AUTO_SORT           = 1 <<  5,
  GTK_PANDA_CLIST_AUTO_RESIZE_BLOCKED = 1 <<  6,
  GTK_PANDA_CLIST_REORDERABLE         = 1 <<  7,
  GTK_PANDA_CLIST_USE_DRAG_ICONS      = 1 <<  8,
  GTK_PANDA_CLIST_DRAW_DRAG_LINE      = 1 <<  9,
  GTK_PANDA_CLIST_DRAW_DRAG_RECT      = 1 << 10
}; 

/* cell types */
typedef enum
{
  GTK_PANDA_CELL_EMPTY,
  GTK_PANDA_CELL_TEXT,
  GTK_PANDA_CELL_PIXMAP,
  GTK_PANDA_CELL_PIXTEXT,
  GTK_PANDA_CELL_WIDGET
} GtkPandaCellType;

typedef enum
{
  GTK_PANDA_CLIST_DRAG_NONE,
  GTK_PANDA_CLIST_DRAG_BEFORE,
  GTK_PANDA_CLIST_DRAG_INTO,
  GTK_PANDA_CLIST_DRAG_AFTER
} GtkPandaCListDragPos;

typedef enum
{
  GTK_PANDA_BUTTON_IGNORED = 0,
  GTK_PANDA_BUTTON_SELECTS = 1 << 0,
  GTK_PANDA_BUTTON_DRAGS   = 1 << 1,
  GTK_PANDA_BUTTON_EXPANDS = 1 << 2
} GtkPandaButtonAction;

#define GTK_PANDA_TYPE_CLIST            (gtkpanda_clist_get_type ())
#define GTK_PANDA_CLIST(obj)            (GTK_CHECK_CAST ((obj), GTK_PANDA_TYPE_CLIST, GtkPandaCList))
#define GTK_PANDA_CLIST_CLASS(klass)    (GTK_CHECK_CLASS_CAST ((klass), GTK_PANDA_TYPE_CLIST, GtkPandaCListClass))
#define GTK_PANDA_IS_CLIST(obj)         (GTK_CHECK_TYPE ((obj), GTK_PANDA_TYPE_CLIST))
#define GTK_PANDA_IS_CLIST_CLASS(klass) (GTK_CHECK_CLASS_TYPE ((klass), GTK_PANDA_TYPE_CLIST))

#define GTK_PANDA_CLIST_FLAGS(clist)             (GTK_PANDA_CLIST (clist)->flags)
#define GTK_PANDA_CLIST_SET_FLAG(clist,flag)     (GTK_PANDA_CLIST_FLAGS (clist) |= (GTK_ ## flag))
#define GTK_PANDA_CLIST_UNSET_FLAG(clist,flag)   (GTK_PANDA_CLIST_FLAGS (clist) &= ~(GTK_ ## flag))

#define GTK_PANDA_CLIST_IN_DRAG(clist)           (GTK_PANDA_CLIST_FLAGS (clist) & GTK_PANDA_CLIST_IN_DRAG)
#define GTK_PANDA_CLIST_ROW_HEIGHT_SET(clist)    (GTK_PANDA_CLIST_FLAGS (clist) & GTK_PANDA_CLIST_ROW_HEIGHT_SET)
#define GTK_PANDA_CLIST_SHOW_TITLES(clist)       (GTK_PANDA_CLIST_FLAGS (clist) & GTK_PANDA_CLIST_SHOW_TITLES)
#define GTK_PANDA_CLIST_CHILD_HAS_FOCUS(clist)   (GTK_PANDA_CLIST_FLAGS (clist) & GTK_PANDA_CLIST_CHILD_HAS_FOCUS)
#define GTK_PANDA_CLIST_ADD_MODE(clist)          (GTK_PANDA_CLIST_FLAGS (clist) & GTK_PANDA_CLIST_ADD_MODE)
#define GTK_PANDA_CLIST_AUTO_SORT(clist)         (GTK_PANDA_CLIST_FLAGS (clist) & GTK_PANDA_CLIST_AUTO_SORT)
#define GTK_PANDA_CLIST_AUTO_RESIZE_BLOCKED(clist) (GTK_PANDA_CLIST_FLAGS (clist) & GTK_PANDA_CLIST_AUTO_RESIZE_BLOCKED)
#define GTK_PANDA_CLIST_REORDERABLE(clist)       (GTK_PANDA_CLIST_FLAGS (clist) & GTK_PANDA_CLIST_REORDERABLE)
#define GTK_PANDA_CLIST_USE_DRAG_ICONS(clist)    (GTK_PANDA_CLIST_FLAGS (clist) & GTK_PANDA_CLIST_USE_DRAG_ICONS)
#define GTK_PANDA_CLIST_DRAW_DRAG_LINE(clist)    (GTK_PANDA_CLIST_FLAGS (clist) & GTK_PANDA_CLIST_DRAW_DRAG_LINE)
#define GTK_PANDA_CLIST_DRAW_DRAG_RECT(clist)    (GTK_PANDA_CLIST_FLAGS (clist) & GTK_PANDA_CLIST_DRAW_DRAG_RECT)

#define GTK_PANDA_CLIST_ROW(_glist_) ((GtkPandaCListRow *)((_glist_)->data))

/* pointer casting for cells */
#define GTK_PANDA_CELL_TEXT(cell)     (((GtkPandaCellText *) &(cell)))
#define GTK_PANDA_CELL_PIXMAP(cell)   (((GtkPandaCellPixmap *) &(cell)))
#define GTK_PANDA_CELL_PIXTEXT(cell)  (((GtkPandaCellPixText *) &(cell)))
#define GTK_PANDA_CELL_WIDGET(cell)   (((GtkPandaCellWidget *) &(cell)))

typedef struct _GtkPandaCList GtkPandaCList;
typedef struct _GtkPandaCListClass GtkPandaCListClass;
typedef struct _GtkPandaCListColumn GtkPandaCListColumn;
typedef struct _GtkPandaCListRow GtkPandaCListRow;

typedef struct _GtkPandaCell GtkPandaCell;
typedef struct _GtkPandaCellText GtkPandaCellText;
typedef struct _GtkPandaCellPixmap GtkPandaCellPixmap;
typedef struct _GtkPandaCellPixText GtkPandaCellPixText;
typedef struct _GtkPandaCellWidget GtkPandaCellWidget;

typedef gint (*GtkPandaCListCompareFunc) (GtkPandaCList     *clist,
				     gconstpointer ptr1,
				     gconstpointer ptr2);

typedef struct _GtkPandaCListCellInfo GtkPandaCListCellInfo;
typedef struct _GtkPandaCListDestInfo GtkPandaCListDestInfo;

struct _GtkPandaCListCellInfo
{
  gint row;
  gint column;
};

struct _GtkPandaCListDestInfo
{
  GtkPandaCListCellInfo cell;
  GtkPandaCListDragPos  insert_pos;
};

struct _GtkPandaCList
{
  GtkContainer container;
  
  guint16 flags;
  
  /* mem chunks */
  GMemChunk *row_mem_chunk;
  GMemChunk *cell_mem_chunk;

  guint freeze_count;
  
  /* allocation rectangle after the conatiner_border_width
   * and the width of the shadow border */
  GdkRectangle internal_allocation;
  
  /* rows */
  gint rows;
  gint row_center_offset;
  gint row_height;
  GList *row_list;
  GList *row_list_end;
  
  /* columns */
  gint columns;
  GdkRectangle column_title_area;
  GdkWindow *title_window;
  
  /* dynamicly allocated array of column structures */
  GtkPandaCListColumn *column;
  
  /* the scrolling window and its height and width to
   * make things a little speedier */
  GdkWindow *clist_window;
  gint clist_window_width;
  gint clist_window_height;
  
  /* offsets for scrolling */
  gint hoffset;
  gint voffset;
  
  /* border shadow style */
  GtkShadowType shadow_type;
  
  /* the list's selection mode (gtkenums.h) */
  GtkSelectionMode selection_mode;
  
  /* list of selected rows */
  GList *selection;
  GList *selection_end;
  
  GList *undo_selection;
  GList *undo_unselection;
  gint undo_anchor;
  
  /* mouse buttons */
  guint8 button_actions[5];

  guint8 drag_button;

  /* dnd */
  GtkPandaCListCellInfo click_cell;

  /* scroll adjustments */
  GtkAdjustment *hadjustment;
  GtkAdjustment *vadjustment;
  
  /* xor GC for the vertical drag line */
  GdkGC *xor_gc;
  
  /* gc for drawing unselected cells */
  GdkGC *fg_gc;
  GdkGC *bg_gc;
  
  /* cursor used to indicate dragging */
  GdkCursor *cursor_drag;
  
  /* the current x-pixel location of the xor-drag line */
  gint x_drag;
  
  /* focus handling */
  gint focus_row;
  
  /* dragging the selection */
  gint anchor;
  GtkStateType anchor_state;
  gint drag_pos;
  gint htimer;
  gint vtimer;
  
  GtkSortType sort_type;
  GtkPandaCListCompareFunc compare;
  gint sort_column;
};

struct _GtkPandaCListClass
{
  GtkContainerClass parent_class;
  
  void  (*set_scroll_adjustments) (GtkPandaCList       *clist,
				   GtkAdjustment  *hadjustment,
				   GtkAdjustment  *vadjustment);
  void   (*refresh)             (GtkPandaCList       *clist);
  void   (*select_row)          (GtkPandaCList       *clist,
				 gint            row,
				 gint            column,
				 GdkEvent       *event);
  void   (*unselect_row)        (GtkPandaCList       *clist,
				 gint            row,
				 gint            column,
				 GdkEvent       *event);
  void   (*row_move)            (GtkPandaCList       *clist,
				 gint            source_row,
				 gint            dest_row);
  void   (*click_column)        (GtkPandaCList       *clist,
				 gint            column);
  void   (*resize_column)       (GtkPandaCList       *clist,
				 gint            column,
                                 gint            width);
  void   (*toggle_focus_row)    (GtkPandaCList       *clist);
  void   (*select_all)          (GtkPandaCList       *clist);
  void   (*unselect_all)        (GtkPandaCList       *clist);
  void   (*undo_selection)      (GtkPandaCList       *clist);
  void   (*start_selection)     (GtkPandaCList       *clist);
  void   (*end_selection)       (GtkPandaCList       *clist);
  void   (*extend_selection)    (GtkPandaCList       *clist,
				 GtkScrollType   scroll_type,
				 gfloat          position,
				 gboolean        auto_start_selection);
  void   (*scroll_horizontal)   (GtkPandaCList       *clist,
				 GtkScrollType   scroll_type,
				 gfloat          position);
  void   (*scroll_vertical)     (GtkPandaCList       *clist,
				 GtkScrollType   scroll_type,
				 gfloat          position);
  void   (*toggle_add_mode)     (GtkPandaCList       *clist);
  void   (*abort_column_resize) (GtkPandaCList       *clist);
  void   (*resync_selection)    (GtkPandaCList       *clist,
				 GdkEvent       *event);
  GList* (*selection_find)      (GtkPandaCList       *clist,
				 gint            row_number,
				 GList          *row_list_element);
  void   (*draw_row)            (GtkPandaCList       *clist,
				 GdkRectangle   *area,
				 gint            row,
				 GtkPandaCListRow    *clist_row);
  void   (*draw_drag_highlight) (GtkPandaCList        *clist,
				 GtkPandaCListRow     *target_row,
				 gint             target_row_number,
				 GtkPandaCListDragPos  drag_pos);
  void   (*clear)               (GtkPandaCList       *clist);
  void   (*fake_unselect_all)   (GtkPandaCList       *clist,
				 gint            row);
  void   (*sort_list)           (GtkPandaCList       *clist);
  gint   (*insert_row)          (GtkPandaCList       *clist,
				 gint            row,
				 gchar          *text[]);
  void   (*remove_row)          (GtkPandaCList       *clist,
				 gint            row);
  void   (*set_cell_contents)   (GtkPandaCList       *clist,
				 GtkPandaCListRow    *clist_row,
				 gint            column,
				 GtkPandaCellType     type,
				 const gchar    *text,
				 guint8          spacing,
				 GdkPixmap      *pixmap,
				 GdkBitmap      *mask);
  void   (*cell_size_request)   (GtkPandaCList       *clist,
				 GtkPandaCListRow    *clist_row,
				 gint            column,
				 GtkRequisition *requisition);

};

struct _GtkPandaCListColumn
{
  gchar *title;
  GdkRectangle area;
  
  GtkWidget *button;
  GdkWindow *window;
  
  gint width;
  gint min_width;
  gint max_width;
  GtkJustification justification;
  
  guint visible        : 1;  
  guint width_set      : 1;
  guint resizeable     : 1;
  guint auto_resize    : 1;
  guint button_passive : 1;
};

struct _GtkPandaCListRow
{
  GtkPandaCell *cell;
  GtkStateType state;
  
  GdkColor foreground;
  GdkColor background;
  
  GtkStyle *style;

  gpointer data;
  GtkDestroyNotify destroy;
  
  guint fg_set     : 1;
  guint bg_set     : 1;
  guint selectable : 1;
};

/* Cell Structures */
struct _GtkPandaCellText
{
  GtkPandaCellType type;
  
  gint16 vertical;
  gint16 horizontal;
  
  GtkStyle *style;

  gchar *text;
};

struct _GtkPandaCellPixmap
{
  GtkPandaCellType type;
  
  gint16 vertical;
  gint16 horizontal;
  
  GtkStyle *style;

  GdkPixmap *pixmap;
  GdkBitmap *mask;
};

struct _GtkPandaCellPixText
{
  GtkPandaCellType type;
  
  gint16 vertical;
  gint16 horizontal;
  
  GtkStyle *style;

  gchar *text;
  guint8 spacing;
  GdkPixmap *pixmap;
  GdkBitmap *mask;
};

struct _GtkPandaCellWidget
{
  GtkPandaCellType type;
  
  gint16 vertical;
  gint16 horizontal;
  
  GtkStyle *style;

  GtkWidget *widget;
};

struct _GtkPandaCell
{
  GtkPandaCellType type;
  
  gint16 vertical;
  gint16 horizontal;
  
  GtkStyle *style;

  union {
    gchar *text;
    
    struct {
      GdkPixmap *pixmap;
      GdkBitmap *mask;
    } pm;
    
    struct {
      gchar *text;
      guint8 spacing;
      GdkPixmap *pixmap;
      GdkBitmap *mask;
    } pt;
    
    GtkWidget *widget;
  } u;
};

GtkType gtkpanda_clist_get_type (void);

/* constructors useful for gtk-- wrappers */
void gtkpanda_clist_construct (GtkPandaCList *clist,
			  gint      columns,
			  gchar    *titles[]);

/* create a new GtkPandaCList */
GtkWidget* gtkpanda_clist_new             (gint   columns);
GtkWidget* gtkpanda_clist_new_with_titles (gint   columns,
				      gchar *titles[]);

/* set adjustments of clist */
void gtkpanda_clist_set_hadjustment (GtkPandaCList      *clist,
				GtkAdjustment *adjustment);
void gtkpanda_clist_set_vadjustment (GtkPandaCList      *clist,
				GtkAdjustment *adjustment);

/* get adjustments of clist */
GtkAdjustment* gtkpanda_clist_get_hadjustment (GtkPandaCList *clist);
GtkAdjustment* gtkpanda_clist_get_vadjustment (GtkPandaCList *clist);

/* set the border style of the clist */
void gtkpanda_clist_set_shadow_type (GtkPandaCList      *clist,
				GtkShadowType  type);

/* set the clist's selection mode */
void gtkpanda_clist_set_selection_mode (GtkPandaCList         *clist,
				   GtkSelectionMode  mode);

/* enable clists reorder ability */
void gtkpanda_clist_set_reorderable (GtkPandaCList *clist,
				gboolean  reorderable);
void gtkpanda_clist_set_use_drag_icons (GtkPandaCList *clist,
				   gboolean  use_icons);
void gtkpanda_clist_set_button_actions (GtkPandaCList *clist,
				   guint     button,
				   guint8    button_actions);

/* freeze all visual updates of the list, and then thaw the list after
 * you have made a number of changes and the updates wil occure in a
 * more efficent mannor than if you made them on a unfrozen list
 */
void gtkpanda_clist_freeze (GtkPandaCList *clist);
void gtkpanda_clist_thaw   (GtkPandaCList *clist);

/* show and hide the column title buttons */
void gtkpanda_clist_column_titles_show (GtkPandaCList *clist);
void gtkpanda_clist_column_titles_hide (GtkPandaCList *clist);

/* set the column title to be a active title (responds to button presses, 
 * prelights, and grabs keyboard focus), or passive where it acts as just
 * a title
 */
void gtkpanda_clist_column_title_active   (GtkPandaCList *clist,
				      gint      column);
void gtkpanda_clist_column_title_passive  (GtkPandaCList *clist,
				      gint      column);
void gtkpanda_clist_column_titles_active  (GtkPandaCList *clist);
void gtkpanda_clist_column_titles_passive (GtkPandaCList *clist);

/* set the title in the column title button */
void gtkpanda_clist_set_column_title (GtkPandaCList    *clist,
				 gint         column,
				 const gchar *title);

/* returns the title of column. Returns NULL if title is not set */
gchar * gtkpanda_clist_get_column_title (GtkPandaCList *clist,
				    gint      column);

/* set a widget instead of a title for the column title button */
void gtkpanda_clist_set_column_widget (GtkPandaCList  *clist,
				  gint       column,
				  GtkWidget *widget);

/* returns the column widget */
GtkWidget * gtkpanda_clist_get_column_widget (GtkPandaCList *clist,
					 gint      column);

/* set the justification on a column */
void gtkpanda_clist_set_column_justification (GtkPandaCList         *clist,
					 gint              column,
					 GtkJustification  justification);

/* set visibility of a column */
void gtkpanda_clist_set_column_visibility (GtkPandaCList *clist,
				      gint      column,
				      gboolean  visible);

/* enable/disable column resize operations by mouse */
void gtkpanda_clist_set_column_resizeable (GtkPandaCList *clist,
				      gint      column,
				      gboolean  resizeable);

/* resize column automatically to its optimal width */
void gtkpanda_clist_set_column_auto_resize (GtkPandaCList *clist,
				       gint      column,
				       gboolean  auto_resize);

gint gtkpanda_clist_columns_autosize (GtkPandaCList *clist);

/* return the optimal column width, i.e. maximum of all cell widths */
gint gtkpanda_clist_optimal_column_width (GtkPandaCList *clist,
				     gint      column);

/* set the pixel width of a column; this is a necessary step in
 * creating a CList because otherwise the column width is chozen from
 * the width of the column title, which will never be right
 */
void gtkpanda_clist_set_column_width (GtkPandaCList *clist,
				 gint      column,
				 gint      width);

/* set column minimum/maximum width. min/max_width < 0 => no restriction */
void gtkpanda_clist_set_column_min_width (GtkPandaCList *clist,
				     gint      column,
				     gint      min_width);
void gtkpanda_clist_set_column_max_width (GtkPandaCList *clist,
				     gint      column,
				     gint      max_width);

/* change the height of the rows, the default (height=0) is
 * the hight of the current font.
 */
void gtkpanda_clist_set_row_height (GtkPandaCList *clist,
			       guint     height);

/* scroll the viewing area of the list to the given column and row;
 * row_align and col_align are between 0-1 representing the location the
 * row should appear on the screnn, 0.0 being top or left, 1.0 being
 * bottom or right; if row or column is -1 then then there is no change
 */
void gtkpanda_clist_moveto (GtkPandaCList *clist,
		       gint      row,
		       gint      column,
		       gfloat    row_align,
		       gfloat    col_align);

/* returns whether the row is visible */
GtkVisibility gtkpanda_clist_row_is_visible (GtkPandaCList *clist,
					gint      row);

/* returns the cell type */
GtkPandaCellType gtkpanda_clist_get_cell_type (GtkPandaCList *clist,
				     gint      row,
				     gint      column);

/* sets a given cell's text, replacing its current contents */
void gtkpanda_clist_set_text (GtkPandaCList    *clist,
			 gint         row,
			 gint         column,
			 const gchar *text);

/* for the "get" functions, any of the return pointer can be
 * NULL if you are not interested
 */
gint gtkpanda_clist_get_text (GtkPandaCList  *clist,
			 gint       row,
			 gint       column,
			 gchar    **text);

/* sets a given cell's pixmap, replacing its current contents */
void gtkpanda_clist_set_pixmap (GtkPandaCList  *clist,
			   gint       row,
			   gint       column,
			   GdkPixmap *pixmap,
			   GdkBitmap *mask);

gint gtkpanda_clist_get_pixmap (GtkPandaCList   *clist,
			   gint        row,
			   gint        column,
			   GdkPixmap **pixmap,
			   GdkBitmap **mask);

/* sets a given cell's pixmap and text, replacing its current contents */
void gtkpanda_clist_set_pixtext (GtkPandaCList    *clist,
			    gint         row,
			    gint         column,
			    const gchar *text,
			    guint8       spacing,
			    GdkPixmap   *pixmap,
			    GdkBitmap   *mask);

gint gtkpanda_clist_get_pixtext (GtkPandaCList   *clist,
			    gint        row,
			    gint        column,
			    gchar     **text,
			    guint8     *spacing,
			    GdkPixmap **pixmap,
			    GdkBitmap **mask);

/* sets the foreground color of a row, the color must already
 * be allocated
 */
void gtkpanda_clist_set_foreground (GtkPandaCList *clist,
			       gint      row,
			       GdkColor *color);

/* sets the background color of a row, the color must already
 * be allocated
 */
void gtkpanda_clist_set_background (GtkPandaCList *clist,
			       gint      row,
			       GdkColor *color);

/* set / get cell styles */
void gtkpanda_clist_set_cell_style (GtkPandaCList *clist,
			       gint      row,
			       gint      column,
			       GtkStyle *style);

GtkStyle *gtkpanda_clist_get_cell_style (GtkPandaCList *clist,
				    gint      row,
				    gint      column);

void gtkpanda_clist_set_row_style (GtkPandaCList *clist,
			      gint      row,
			      GtkStyle *style);

GtkStyle *gtkpanda_clist_get_row_style (GtkPandaCList *clist,
				   gint      row);

/* this sets a horizontal and vertical shift for drawing
 * the contents of a cell; it can be positive or negitive;
 * this is particulary useful for indenting items in a column
 */
void gtkpanda_clist_set_shift (GtkPandaCList *clist,
			  gint      row,
			  gint      column,
			  gint      vertical,
			  gint      horizontal);

/* set/get selectable flag of a single row */
void gtkpanda_clist_set_selectable (GtkPandaCList *clist,
			       gint      row,
			       gboolean  selectable);
gboolean gtkpanda_clist_get_selectable (GtkPandaCList *clist,
				   gint      row);

/* prepend/append returns the index of the row you just added,
 * making it easier to append and modify a row
 */
gint gtkpanda_clist_prepend (GtkPandaCList    *clist,
		        gchar       *text[]);
gint gtkpanda_clist_append  (GtkPandaCList    *clist,
			gchar       *text[]);

/* inserts a row at index row and returns the row where it was
 * actually inserted (may be different from "row" in auto_sort mode)
 */
gint gtkpanda_clist_insert (GtkPandaCList    *clist,
		       gint         row,
		       gchar       *text[]);

/* removes row at index row */
void gtkpanda_clist_remove (GtkPandaCList *clist,
		       gint      row);

/* sets a arbitrary data pointer for a given row */
void gtkpanda_clist_set_row_data (GtkPandaCList *clist,
			     gint      row,
			     gpointer  data);

/* sets a data pointer for a given row with destroy notification */
void gtkpanda_clist_set_row_data_full (GtkPandaCList         *clist,
			          gint              row,
			          gpointer          data,
				  GtkDestroyNotify  destroy);

/* returns the data set for a row */
gpointer gtkpanda_clist_get_row_data (GtkPandaCList *clist,
				 gint      row);

/* givin a data pointer, find the first (and hopefully only!)
 * row that points to that data, or -1 if none do
 */
gint gtkpanda_clist_find_row_from_data (GtkPandaCList *clist,
				   gpointer  data);

/* force selection of a row */
void gtkpanda_clist_select_row (GtkPandaCList *clist,
			   gint      row,
			   gint      column);

/* force unselection of a row */
void gtkpanda_clist_unselect_row (GtkPandaCList *clist,
			     gint      row,
			     gint      column);

/* undo the last select/unselect operation */
void gtkpanda_clist_undo_selection (GtkPandaCList *clist);

/* clear the entire list -- this is much faster than removing
 * each item with gtkpanda_clist_remove
 */
void gtkpanda_clist_clear (GtkPandaCList *clist);

/* return the row column corresponding to the x and y coordinates,
 * the returned values are only valid if the x and y coordinates
 * are respectively to a window == clist->clist_window
 */
gint gtkpanda_clist_get_selection_info (GtkPandaCList *clist,
			     	   gint      x,
			     	   gint      y,
			     	   gint     *row,
			     	   gint     *column);

/* in multiple or extended mode, select all rows */
void gtkpanda_clist_select_all (GtkPandaCList *clist);

/* in all modes except browse mode, deselect all rows */
void gtkpanda_clist_unselect_all (GtkPandaCList *clist);

/* swap the position of two rows */
void gtkpanda_clist_swap_rows (GtkPandaCList *clist,
			  gint      row1,
			  gint      row2);

/* move row from source_row position to dest_row position */
void gtkpanda_clist_row_move (GtkPandaCList *clist,
			 gint      source_row,
			 gint      dest_row);

/* sets a compare function different to the default */
void gtkpanda_clist_set_compare_func (GtkPandaCList            *clist,
				 GtkPandaCListCompareFunc  cmp_func);

/* the column to sort by */
void gtkpanda_clist_set_sort_column (GtkPandaCList *clist,
				gint      column);

/* how to sort : ascending or descending */
void gtkpanda_clist_set_sort_type (GtkPandaCList    *clist,
			      GtkSortType  sort_type);

/* sort the list with the current compare function */
void gtkpanda_clist_sort (GtkPandaCList *clist);

/* Automatically sort upon insertion */
void gtkpanda_clist_set_auto_sort (GtkPandaCList *clist,
			      gboolean  auto_sort);


#ifdef __cplusplus
}
#endif				/* __cplusplus */

#endif				/* __GTK_PANDA_CLIST_H__ */
