/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GTK+ Team and others 1997-1999.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
 */

#define USE_XIM

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk-xmhtml/gtk-xmhtml.h>
#include <ghttp.h>
#include <pthread.h>
#include "gtkpandahtml.h"

static void gtk_panda_html_class_init    (GtkPandaHTMLClass *klass);
static void gtk_panda_html_init          (GtkPandaHTML     *html);

GtkType
gtk_panda_html_get_type (void)
{
  static GtkType type = 0;

  if (!type)
    {
      static const GtkTypeInfo info =
      {
	"GtkPandaHTML",
	sizeof (GtkPandaHTML),
	sizeof (GtkPandaHTMLClass),
	(GtkClassInitFunc) gtk_panda_html_class_init,
	(GtkObjectInitFunc) gtk_panda_html_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL,
      };

      type = gtk_type_unique (gtk_xmhtml_get_type (), &info);
    }

  return type;
}

static void
gtk_panda_html_class_init (GtkPandaHTMLClass *class)
{
}

static void
gtk_panda_html_init (GtkPandaHTML *html)
{
  html->uri = NULL;
}

GtkWidget*
gtk_panda_html_new (void)
{
  return GTK_WIDGET (gtk_type_new (GTK_PANDA_TYPE_HTML));
}

char *
gtk_panda_html_get_uri (GtkPandaHTML *html)
{
  return html->uri;
}

static pthread_t loading_thread;

static struct html_data {
  GtkXmHTML *html;
  char *source;
} html_data;

static int
set_page (gpointer data)
{
  struct html_data *p = data;
  gtk_xmhtml_source (p->html, p->source);
  free (p->source);
  return FALSE;
}

static void *
start_loading (void *data)
{
  ghttp_request *rq;
  GtkPandaHTML *html = (GtkPandaHTML *) data;

  /* Download */
  rq = ghttp_request_new ();
  html_data.html = GTK_XMHTML (html);
  if (ghttp_set_uri (rq, html->uri) != -1
      && ghttp_prepare (rq) != -1
      && ghttp_process (rq) == ghttp_done)
    {
      html_data.source = strdup (ghttp_get_body (rq));
      gtk_idle_add (set_page, &html_data);
    }
  else
    {
      char msg[BUFSIZ];
      sprintf (msg, "<body>Cannot connect to <tt>%s</tt></body>", html->uri);
      html_data.source = strdup (msg);
      gtk_idle_add (set_page, &html_data);
    }
  ghttp_clean (rq);
  return NULL;
}

void
gtk_panda_html_set_uri (GtkPandaHTML *html, const char *uri)
{
  char msg[BUFSIZ];

  /* Save URI */
  if (html->uri)
    free (html->uri);
  html->uri = strdup (uri);

  if (ghttp_uri_validate (html->uri) == -1)
    return;

  /* Start loading */
  sprintf (msg, "<body>Connecting to <tt>%s...</tt></body>", uri);
  gtk_xmhtml_source (GTK_XMHTML (html), msg);
  pthread_create (&loading_thread, NULL, start_loading, html);
}
