/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GTK+ Team and others 1997-1999.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
 */

#define USE_XIM

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <gtk/gtktext.h>
#include <gtk/gtksignal.h>
#include <gdk/gdkprivate.h>
#include <X11/Xlib.h>
#include "gtkpandatext.h"

static void gtk_panda_text_class_init    (GtkPandaTextClass *klass);
static void gtk_panda_text_init          (GtkPandaText     *text);

static gint gtk_panda_text_focus_in      (GtkWidget         *widget,
					  GdkEventFocus     *event);

static GtkTextClass *parent_class = NULL;

GtkType
gtk_panda_text_get_type (void)
{
  static GtkType type = 0;

  if (!type)
    {
      static const GtkTypeInfo info =
      {
	"GtkPandaText",
	sizeof (GtkPandaText),
	sizeof (GtkPandaTextClass),
	(GtkClassInitFunc) gtk_panda_text_class_init,
	(GtkObjectInitFunc) gtk_panda_text_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL,
      };

      type = gtk_type_unique (GTK_TYPE_TEXT, &info);
    }

  return type;
}

static void
gtk_panda_text_class_init (GtkPandaTextClass *class)
{
  GtkWidgetClass *widget_class;

  widget_class = (GtkWidgetClass*) class;
  parent_class = gtk_type_class (GTK_TYPE_TEXT);

  widget_class->focus_in_event = gtk_panda_text_focus_in;
}

static void
gtk_panda_text_init (GtkPandaText *text)
{
  text->xim_enabled = FALSE;
}

GtkWidget*
gtk_panda_text_new (GtkAdjustment *hadj,
		    GtkAdjustment *vadj)
{
  GtkWidget *text;

  if (hadj)
    g_return_val_if_fail (GTK_IS_ADJUSTMENT (hadj), NULL);
  if (vadj)
    g_return_val_if_fail (GTK_IS_ADJUSTMENT (vadj), NULL);

  text = gtk_widget_new (GTK_PANDA_TYPE_TEXT,
			 "hadjustment", hadj,
			 "vadjustment", vadj,
			 NULL);

  return text;
}

void
gtk_panda_text_set_xim_enabled (GtkPandaText *text, gboolean flag)
{
  text->xim_enabled = flag;
}

static gint
gtk_panda_text_focus_in (GtkWidget     *widget,
			 GdkEventFocus *event)
{
  GtkText *text;
  GtkEditable *editable;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_TEXT (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  text = GTK_TEXT (widget);
  editable = GTK_EDITABLE (widget);

  if (GTK_WIDGET_CLASS (parent_class)->focus_in_event)
    (* GTK_WIDGET_CLASS (parent_class)->focus_in_event) (widget, event);

#ifdef USE_XIM
  if (GTK_PANDA_TEXT (widget)->xim_enabled && gdk_xim_ic && gdk_xim_ic->xic)
    {
      XVaNestedList *preedit_attr =
	XVaCreateNestedList (0, XNPreeditState, XIMPreeditEnable, NULL);
      XSetICValues (gdk_xim_ic->xic, XNPreeditAttributes, preedit_attr, NULL);
      XFree (preedit_attr);
    }
#endif

  return FALSE;
}
