/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GTK+ Team and others 1997-1999.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
 */

#include <gtk/gtk.h>
#include "gtkpandatimer.h"

enum {
  TIMEOUT,
  LAST_SIGNAL
};

static void gtk_panda_timer_class_init    (GtkPandaTimerClass *klass);
static void gtk_panda_timer_init          (GtkPandaTimer      *timer);

static guint timer_signals[LAST_SIGNAL] = { 0 };

GtkType
gtk_panda_timer_get_type (void)
{
  static GtkType type = 0;

  if (!type)
    {
      static const GtkTypeInfo info =
      {
	"GtkPandaTimer",
	sizeof (GtkPandaTimer),
	sizeof (GtkPandaTimerClass),
	(GtkClassInitFunc) gtk_panda_timer_class_init,
	(GtkObjectInitFunc) gtk_panda_timer_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL,
      };

      type = gtk_type_unique (GTK_TYPE_WIDGET, &info);
    }

  return type;
}

static void
gtk_panda_timer_class_init (GtkPandaTimerClass *class)
{
  GtkObjectClass *object_class;

  object_class = (GtkObjectClass *) class;

  timer_signals[TIMEOUT] =
    gtk_signal_new ("timeout",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GtkPandaTimerClass, timeout),
		    gtk_marshal_NONE__NONE,
		    GTK_TYPE_NONE, 0);

  gtk_object_class_add_signals (object_class, timer_signals, LAST_SIGNAL);
}

static void
gtk_panda_timer_init (GtkPandaTimer *timer)
{
  GTK_WIDGET_SET_FLAGS (timer, GTK_NO_WINDOW);
}

GtkWidget*
gtk_panda_timer_new (void)
{
  GtkWidget *timer;

  timer = gtk_type_new (GTK_PANDA_TYPE_TIMER);
  GTK_PANDA_TIMER (timer)->tag = 0;
  gtk_panda_timer_set (GTK_PANDA_TIMER (timer), 0);

  return timer;
}

void
gtk_panda_timer_set (GtkPandaTimer *timer, guint32 duration)
{
  timer->duration = duration * 1000;
}

static gint
timeout_handler (gpointer data)
{
  GtkPandaTimer *timer = data;

  if (GTK_WIDGET_VISIBLE (GTK_WIDGET (timer)))
    gtk_signal_emit (GTK_OBJECT (timer), timer_signals[TIMEOUT]);

  return FALSE;
}

void
gtk_panda_timer_reset (GtkPandaTimer *timer)
{
  if (timer->tag != 0)
    gtk_timeout_remove (timer->tag);

  if (timer->duration > 0)
    timer->tag = gtk_timeout_add (timer->duration, timeout_handler, timer);
  else
    timer->tag = 0;
}
