/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 * Copyright (C) 2001 AXE, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Based on gtk/pandatree.h
 */

/*
 * Modified by the GTK+ Team and others 1997-1999.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
 */

#ifndef __GTKPANDA_TREE_H__
#define __GTK_PANDA_TREE_H__

/* set this flag to enable tree debugging output */
/* #define TREE_DEBUG */

#include <gdk/gdk.h>
#include <gtk/gtkcontainer.h>


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


#define GTK_PANDA_TYPE_TREE                  (gtkpanda_tree_get_type ())
#define GTK_PANDA_TREE(obj)                  (GTK_CHECK_CAST ((obj), GTK_PANDA_TYPE_TREE, GtkPandaTree))
#define GTK_PANDA_TREE_CLASS(klass)          (GTK_CHECK_CLASS_CAST ((klass), GTK_PANDA_TYPE_TREE, GtkPandaTreeClass))
#define GTK_PANDA_IS_TREE(obj)               (GTK_CHECK_TYPE ((obj), GTK_PANDA_TYPE_TREE))
#define GTK_PANDA_IS_TREE_CLASS(klass)       (GTK_CHECK_CLASS_TYPE ((klass), GTK_PANDA_TYPE_TREE))

#define GTK_PANDA_IS_ROOT_TREE(obj)   ((GtkObject*) GTK_PANDA_TREE(obj)->root_tree == (GtkObject*)obj)
#define GTK_PANDA_TREE_ROOT_TREE(obj) (GTK_PANDA_TREE(obj)->root_tree ? GTK_PANDA_TREE(obj)->root_tree : GTK_PANDA_TREE(obj))
#define GTK_PANDA_TREE_SELECTION(obj) (GTK_PANDA_TREE_ROOT_TREE(obj)->selection)

typedef enum 
{
  GTK_PANDA_TREE_VIEW_LINE,  /* default view mode */
  GTK_PANDA_TREE_VIEW_ITEM
} GtkPandaTreeViewMode;

typedef struct _GtkPandaTree       GtkPandaTree;
typedef struct _GtkPandaTreeClass  GtkPandaTreeClass;

struct _GtkPandaTree
{
  GtkContainer container;
  
  GList *children;
  
  GtkPandaTree* root_tree; /* owner of selection list */
  GtkWidget* tree_owner;
  GList *selection;
  guint level;
  guint indent_value;
  guint current_indent;
  guint selection_mode : 2;
  guint view_mode : 1;
  guint view_line : 1;
};

struct _GtkPandaTreeClass
{
  GtkContainerClass parent_class;
  
  void (* selection_changed) (GtkPandaTree   *tree);
  void (* select_child)      (GtkPandaTree   *tree,
			      GtkWidget *child);
  void (* unselect_child)    (GtkPandaTree   *tree,
			      GtkWidget *child);
};


GtkType    gtkpanda_tree_get_type           (void);
GtkWidget* gtkpanda_tree_new                (void);
void       gtkpanda_tree_append             (GtkPandaTree          *tree,
				        GtkWidget        *tree_item);
void       gtkpanda_tree_prepend            (GtkPandaTree          *tree,
				        GtkWidget        *tree_item);
void       gtkpanda_tree_insert             (GtkPandaTree          *tree,
				        GtkWidget        *tree_item,
				        gint              position);
void       gtkpanda_tree_remove_items       (GtkPandaTree          *tree,
				        GList            *items);
void       gtkpanda_tree_clear_items        (GtkPandaTree          *tree,
				        gint              start,
				        gint              end);
void       gtkpanda_tree_select_item        (GtkPandaTree          *tree,
				        gint              item);
void       gtkpanda_tree_unselect_item      (GtkPandaTree          *tree,
				        gint              item);
void       gtkpanda_tree_select_child       (GtkPandaTree          *tree,
				        GtkWidget        *tree_item);
void       gtkpanda_tree_unselect_child     (GtkPandaTree          *tree,
				        GtkWidget        *tree_item);
gint       gtkpanda_tree_child_position     (GtkPandaTree          *tree,
				        GtkWidget        *child);
void       gtkpanda_tree_set_selection_mode (GtkPandaTree          *tree,
				        GtkSelectionMode  mode);
void       gtkpanda_tree_set_view_mode      (GtkPandaTree          *tree,
				        GtkPandaTreeViewMode   mode); 
void       gtkpanda_tree_set_view_lines     (GtkPandaTree          *tree,
					guint            flag);

/* deprecated function, use gtk_container_remove instead.
 */
void       gtkpanda_tree_remove_item        (GtkPandaTree          *tree,
				        GtkWidget        *child);


#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif /* __GTK_PANDA_TREE_H__ */
