/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 * Copyright (C) 2001 AXE, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Based on gtk/pandatree.h
 */

/*
 * Modified by the GTK+ Team and others 1997-1999.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
 */

#ifndef __GTK_PANDA_TREE_ITEM_H__
#define __GTK_PANDA_TREE_ITEM_H__


#include <gdk/gdk.h>
#include <gtk/gtkitem.h>


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


#define GTK_PANDA_TYPE_TREE_ITEM              (gtkpanda_tree_item_get_type ())
#define GTK_PANDA_TREE_ITEM(obj)              (GTK_CHECK_CAST ((obj), GTK_PANDA_TYPE_TREE_ITEM, GtkPandaTreeItem))
#define GTK_PANDA_TREE_ITEM_CLASS(klass)      (GTK_CHECK_CLASS_CAST ((klass), GTK_PANDA_TYPE_TREE_ITEM, GtkPandaTreeItemClass))
#define GTK_PANDA_IS_TREE_ITEM(obj)           (GTK_CHECK_TYPE ((obj), GTK_PANDA_TYPE_TREE_ITEM))
#define GTK_PANDA_IS_TREE_ITEM_CLASS(klass)   (GTK_CHECK_CLASS_TYPE ((klass), GTK_PANDA_TYPE_TREE_ITEM))

#define GTK_PANDA_TREE_ITEM_SUBTREE(obj)      (GTK_PANDA_TREE_ITEM(obj)->subtree)


typedef struct _GtkPandaTreeItem       GtkPandaTreeItem;
typedef struct _GtkPandaTreeItemClass  GtkPandaTreeItemClass;

struct _GtkPandaTreeItem
{
  GtkItem item;

  GtkWidget *subtree;
  GtkWidget *pixmaps_box;
  GtkWidget *plus_pix_widget, *minus_pix_widget;

  GList *pixmaps;		/* pixmap node for this items color depth */

  guint expanded : 1;
};

struct _GtkPandaTreeItemClass
{
  GtkItemClass parent_class;

  void (* expand)   (GtkPandaTreeItem *tree_item);
  void (* collapse) (GtkPandaTreeItem *tree_item);
};


GtkType    gtkpanda_tree_item_get_type       (void);
GtkWidget* gtkpanda_tree_item_new            (void);
GtkWidget* gtkpanda_tree_item_new_with_label (const gchar *label);
void       gtkpanda_tree_item_set_subtree    (GtkPandaTreeItem *tree_item,
					 GtkWidget   *subtree);
void       gtkpanda_tree_item_remove_subtree (GtkPandaTreeItem *tree_item);
void       gtkpanda_tree_item_select         (GtkPandaTreeItem *tree_item);
void       gtkpanda_tree_item_deselect       (GtkPandaTreeItem *tree_item);
void       gtkpanda_tree_item_expand         (GtkPandaTreeItem *tree_item);
void       gtkpanda_tree_item_collapse       (GtkPandaTreeItem *tree_item);


#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif /* __GTK_PANDA_TREE_ITEM_H__ */
